#include "Sequencer.h"
#include <MainLoop.h>
#include "Static.h"

Sequencer::Sequencer() :
  Task("Sequencer"),
  m_idx(0),
  m_sceneBeforeIdle(-1)
{
  state = Task::Running;
}

void
Sequencer::Scene::start()
{
    for(const auto& pattern : patterns) {
        //pattern.getBytes(namebuf, sizeof(namebuf));
        Log.info("sequencer: Starting pattern task %s", pattern.c_str());
        MainLoop::instance()->dispatch(InputEvent{InputEvent::StartThing, pattern.c_str()});
    }
}

void
Sequencer::Scene::stop()
{
    for(const auto& pattern : patterns) {
        Log.info("sequencer: Stopping pattern task %s", pattern.c_str());
        MainLoop::instance()->dispatch(InputEvent{InputEvent::StopThing, pattern.c_str()});
    }
}

void
Sequencer::loop()
{
  ConfigTaskMixin::loop();
}

const char*
Sequencer::currentSceneName()
{
    return m_scenes[m_idx].name.c_str();
}

const std::vector<Sequencer::Scene>
Sequencer::scenes() const
{
    return m_scenes;
}

void
Sequencer::onStart()
{
}

void
Sequencer::handleEvent(const InputEvent& evt)
{
  if (!m_scenes.empty()) {
    if (evt.intent == InputEvent::ReadyToRoll) {
      Log.info("sequencer: Activating startup scene");
      if (!changeToScene("Start")) {
        Log.info("sequencer: Start scene not found, falling back to first scene");
        changeToScene(m_scenes[0].name.c_str());
      }
    } else if (evt.intent == InputEvent::SetScene && m_scenes[m_idx].name == evt.asString()) {
        return;
    } else if (evt.intent == InputEvent::SetScene) {
      changeToScene(evt.asString());
    } else if (evt.intent == InputEvent::IdleStart) {
      Log.info("sequencer: Activating idle scene");
      m_sceneBeforeIdle = m_idx;
      MainLoop::instance()->dispatch(InputEvent{InputEvent::SetScene, "Idle"});
    } else if (evt.intent == InputEvent::IdleStop && m_sceneBeforeIdle >= 0) {
      Log.info("sequencer: Resuming pre-idle scene");
      //changeToScene(m_scenes[m_sceneBeforeIdle].name.c_str());
      String sceneName = m_scenes[m_sceneBeforeIdle].name;
      m_sceneBeforeIdle = -1;
      MainLoop::instance()->dispatch(InputEvent{InputEvent::SetScene, sceneName.c_str()});
    } 
  }
  ConfigTaskMixin::handleEvent(evt);
}

void
Sequencer::handleConfigChange(const Configuration& cfg)
{
  Log.notice("sequencer: Loading scenes");

  JsonObject sceneList = cfg.get("scenes");
  m_idx = 0;
  m_scenes.clear();
  for(JsonPair pair : sceneList) {
    Log.notice("sequencer: Found scene %s", pair.key().c_str());
    std::vector<String> patterns;
    for(const char* taskName : pair.value().as<JsonArray>()) {
      patterns.push_back(taskName);
    }
    m_scenes.push_back(Scene{pair.key().c_str(), patterns});
  }
  if (m_scenes.empty()) {
    Log.warning("sequencer: Zero scenes loaded!");
  }
}

bool
Sequencer::changeToScene(const char* name)
{
    Log.notice("sequencer: Switching scene to %s!", name);

    int newIdx = 0;
    bool found = false;
    for(newIdx = 0; newIdx < m_scenes.size(); newIdx++) {
      if (!strcmp(name, m_scenes[newIdx].name.c_str())) {
          found = true;
          break;
      }
    }

    if (found) {
      m_scenes[m_idx].stop();
      m_idx = newIdx;
      m_scenes[m_idx].start();
      return true;
    } else {
      Log.notice("sequencer: Scene not found!");
      return false;
    }
}

void
Sequencer::doScenes(Args& args, Print& out)
{
  out.println("Available scenes: ");
  for (auto scene : scenes()) {
    out.println(scene.name.c_str());
  }
}

static String s;

void
Sequencer::doScene(Args& args, Print& out)
{
  String sceneName = args[1];
  MainLoop::instance()->dispatch(InputEvent{InputEvent::SetScene, sceneName.c_str()});
}

const std::vector<Command>&
Sequencer::commands() const
{
  static const std::vector<Command> _commands = {
    {"scene", &Sequencer::doScene},
    {"scenes", &Sequencer::doScenes}
  };
  return _commands;
}

STATIC_ALLOC(Sequencer);
STATIC_TASK(Sequencer);
